# -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
import orjson


async def get_redis_message(message_byte: bytes) -> dict:
    """ """

    if message_byte and message_byte["type"] == "message":

        message_byte_data = orjson.loads(message_byte["data"])

        params = message_byte_data["params"]

        return dict(
            data=params["data"],
            channel=params["channel"],
            message_all=message_byte_data,
        )

    else:

        return dict(
            data=[],
            channel=[],
            message_all=[],
        )



# -*- coding: utf-8 -*-

import calendar
import time
from datetime import datetime, timedelta, timezone


def get_current_local_date_time():
    """ """
    return datetime.now()


def translate_current_local_date_time_to_utc():
    """ """
    return (
        get_current_local_date_time()
        .astimezone()
        .astimezone(timezone.utc)
        .replace(tzinfo=None)
    )


def time_format_standardization(time_input, time_format: str = "%Y-%m-%d %H:%M:%S.%f"):
    """
    Standardize the time format.
    """
    # If the input is a string, then convert it to datetime format.
    strptime = (
        None
        if isinstance(time_input, str) == False
        else datetime.strptime(time_input, time_format)
    )

    # If the input is a datetime, then convert it to string format.
    strftime = (
        None
        if isinstance(time_input, datetime) == False
        else time_input.strftime(time_format)
    )

    # Return a dictionary with two keys: strptime and strftime.
    return {"strptime": strptime, "strftime": strftime}


def convert_time_to_utc(
    transaction_time: str = None, hours_diff_with_utc: float = None
):
    """
    # Mendapatkan waktu UTC/JKT saat ini (now) dengan UTC sebagai patokan

        Args:
            param (None): None

        Return and rtype:
            waktu utc/jkt dalam format yyyy/mm/dd --> dict

        Referensi:
            https://stackoverflow.com/questions/3327946/how-can-i-get-the-current-time-now-in-utc
    """

    utc = translate_current_local_date_time_to_utc()

    if transaction_time != None:
        transaction_time_ = datetime.fromisoformat(transaction_time)
        transaction_time = (
            transaction_time_.astimezone().astimezone(timezone.utc).replace(tzinfo=None)
        )
        utc_f_transaction_time = (
            transaction_time
            + timedelta(hours=0 if hours_diff_with_utc == None else hours_diff_with_utc)
        ).strftime("%Y-%m-%d %H:%M:%S.%f")

    utc_f = time_format_standardization(utc)["strftime"]
    utc_f_jkt = time_format_standardization((utc + timedelta(hours=7)))["strftime"]

    return {
        "utc_now": time_format_standardization(utc_f)["strptime"],
        "jkt_now": time_format_standardization(utc_f_jkt)["strptime"],
        "transaction_time": (
            None
            if transaction_time == None
            else time_format_standardization(utc_f_transaction_time)["strptime"]
        ),
    }


def check_day_name(time: datetime) -> str:
    """
    check day name based on time given
    """
    # time in datetime format
    try:
        return time.strftime("%A")

    # time in text format
    except:
        # convert string to time format
        time_in_time_format = time_format_standardization(time)["strptime"]
        return time_in_time_format.strftime("%A")


def convert_time_to_unix(time) -> int:
    """
    # Get time  (milliseconds since the UNIX epoch)

        Args:
            param1 (Date): Waktu dalam format '%Y-%m-%d %H:%M:%S.%f'
            one minute = 60000

        Return and rtype:
            waktu dalam format UNIX (microseconds) utc/jkt --> int

        Referensi:
            https://stackoverflow.com/questions/41856393/time-data-conversion-missing-microseconds-in-gm-time-python
            https://stackoverflow.com/questions/10611328/parsing-datetime-strings-containing-nanoseconds
            https://stackoverflow.com/questions/58939822/python-does-not-match-format-y-m-dthmsz-f

    """

    try:
        try:
            time_ = 0 if time == 0 else datetime.fromisoformat(time)
            time = 0 if time == 0 else time_.strftime("%Y-%m-%d %H:%M:%S.%f")
            time = 0 if time == 0 else datetime.strptime(time, "%Y-%m-%d %H:%M:%S.%f")
            microsecs = time.microsecond

        except:
            microsecs = time.microsecond  # menarik microsecond untuk dihitung terpisah

    except Exception as error:
        import traceback

        from loguru import logger as log

        log.critical(f"{error}")
        print(traceback.format_exc())

    return int((calendar.timegm(time.timetuple()) * 1000000 + microsecs) / 1000)


def time_delta_between_now_and_transaction_time_both_in_utc(
    transaction_time: str,
) -> float:
    """Menghitung selisih  antara waktu saat ini dengan
    waktu saat transaksi

    """
    none_data = [None, 0, []]

    now_time_utc = convert_time_to_utc()["utc_now"]
    transaction_time_utc = convert_time_to_utc(transaction_time)["transaction_time"]

    # time_delta in seconds
    time_delta = (
        0
        if transaction_time in none_data
        else ((transaction_time_utc - now_time_utc).total_seconds())
    )

    return {
        "seconds": time_delta,
        "hours": time_delta / 3600,
        "days": time_delta / 3600 / 24,
    }


def time_delta_between_two_times(
    time_format, start_time: str, end_time: str = None
) -> float:
    """Menghitung selisih  antara waktu 2 waktu
    time format: utc/unix

    """
    # end time = now
    transaction_time_end_utc = convert_time_to_utc()["utc_now"]
    transaction_time_end_unix = convert_time_to_unix(transaction_time_end_utc)

    if time_format == "utc":
        if end_time != None:
            transaction_time_end_utc = convert_time_to_utc(end_time)["transaction_time"]
            transaction_time_end_unix = convert_time_to_unix(transaction_time_end_utc)

        transaction_time_start_utc = convert_time_to_utc(start_time)["transaction_time"]
        transaction_time_start_unix = convert_time_to_unix(transaction_time_start_utc)
        # transaction_time_end_utc = convert_time_to_utc (end_time)['transaction_time']

        # time_delta in seconds
        time_delta_utc = (
            transaction_time_end_utc - transaction_time_start_utc
        ).total_seconds()
        time_delta_unix = transaction_time_end_unix - transaction_time_start_unix

    if "unix" in time_format:
        seconds_divider = 1000 if "ms" in time_format else 1000
        transaction_time_end_unix = (
            transaction_time_end_unix if end_time == None else end_time
        )

        # time_delta in seconds
        time_delta_unix = transaction_time_end_unix - start_time

    return {
        "seconds": (
            time_delta_unix / seconds_divider
            if "unix" in time_format
            else time_delta_utc
        ),
        "hours": (
            time_delta_unix / 3600 / seconds_divider
            if "unix" in time_format
            else time_delta_utc / 3600
        ),
        "days": (
            time_delta_unix / 3600 / seconds_divider / 24
            if "unix" in time_format
            else time_delta_utc / 3600 / 24
        ),
    }


def check_alarm_clock(
    triggerHour, triggerMinute, isTrigger=False, local_time: str = None
):
    """

    Add an alarm clock to the trading strategy

    Args:
        triggerHour (int)
        triggerMinute (int)

    Returns:
        str

    Example:
        data_original = 'hedgingSpot-open-1671189554374' become 'hedgingSpot'

    Reference:
        https://medium.com/@FMZQuant/add-an-alarm-clock-to-the-trading-strategy-e90e0372405f
    """

    if local_time == "jkt_now":
        current_time = convert_time_to_utc()["jkt_now"]
    else:
        current_time = convert_time_to_utc()["utc_now"]

    t = time.localtime(time.time())
    hour = t.tm_hour
    minute = t.tm_min
    day = t.tm_wday

    nowDay = time.localtime(time.time()).tm_wday

    if day != nowDay:
        isTrigger = False
        nowDay = day

    if isTrigger == False and hour == triggerHour and minute >= triggerMinute:
        isTrigger = True
        return True

    return False


def resampling_time_frame(ohlc_data: list, time_frame: str = "5min"):
    """


    Args:
        triggerHour (int)
        triggerMinute (int)

    Returns:
        str

    Example:
        data_original = 'hedgingSpot-open-1671189554374' become 'hedgingSpot'

    Reference:
        https://stackoverflow.com/questions/69350436/how-do-i-resample-to-5-min-correctly
        https://neel-81024.medium.com/resampling-and-ohlc-in-d967cfafc657
        https://stackoverflow.com/questions/70143660/pandas-ohlc-aggregation-resample-with-time-of-ohlc
        https://stackoverflow.com/questions/36681945/group-dataframe-in-5-minute-intervals
        https://stackoverflow.com/questions/36222928/pandas-ohlc-aggregation-on-ohlc-data
    """
    import pandas as pd

    df = pd.DataFrame(ohlc_data)
    df.set_index(df["Datetime"], inplace=True)
    d = {
        "open": "first",
        "high": "max",
        "low": "min",
        "close": "last",
        "volume": "sum",
    }
    df.resample(time_frame, closed="right", label="right").agg(d)

    return False


def get_now_unix_time() -> int:
    """ """

    now_utc = convert_time_to_utc()["utc_now"]

    return int(convert_time_to_unix(now_utc))



# -*- coding: utf-8 -*-

# built ins
import asyncio

# installed
from loguru import logger as log

from streaming_helper.db_management import sqlite_management as db_mgt
from streaming_helper.channel_management import get_published_messages
from streaming_helper.strategies.deribit.cash_carry import reassigning_labels
from streaming_helper.transaction_management.deribit import cancelling_active_orders
from streaming_helper.data_announcer.deribit import get_instrument_summary
from streaming_helper.channel_management.deribit import subscribing_to_channels
from streaming_helper.utilities import (
    pickling,
    string_modification as str_mod,
    system_tools,
    time_modification as time_mod,
    error_handling,
)


async def relabelling_trades(
    client_id: str,
    client_secret: str,
    client_redis: object,
    config_app: list,
    redis_channels: list,
) -> None:
    """ """

    try:

        # connecting to redis pubsub
        pubsub: object = client_redis.pubsub()

        # subscribe to channels
        await subscribing_to_channels.redis_channels(
            pubsub,
            redis_channels,
            "relabelling",
        )

        # get tradable strategies
        tradable_config_app = config_app["tradable"]

        # get tradable currencies
        currencies = ([o["spot"] for o in tradable_config_app])[0]

        strategy_attributes = config_app["strategies"]

        strategy_attributes_active = [
            o for o in strategy_attributes if o["is_active"] == True
        ]

        active_strategies = [o["strategy_label"] for o in strategy_attributes_active]

        cancellable_strategies = [
            o["strategy_label"]
            for o in strategy_attributes_active
            if o["cancellable"] == True
        ]

        relevant_tables = config_app["relevant_tables"][0]

        order_db_table = relevant_tables["orders_table"]

        settlement_periods = get_settlement_period(strategy_attributes)

        futures_instruments = await get_instrument_summary.get_futures_instruments(
            currencies,
            settlement_periods,
        )

        instrument_attributes_futures_all = futures_instruments["active_futures"]

        # get redis channels
        order_rest_channel: str = redis_channels["order_rest"]
        my_trade_receiving_channel: str = redis_channels["my_trade_receiving"]
        order_update_channel: str = redis_channels["order_cache_updating"]
        portfolio_channel: str = redis_channels["portfolio"]
        sqlite_updating_channel: str = redis_channels["sqlite_record_updating"]
        sub_account_cached_channel: str = redis_channels["sub_account_cache_updating"]

        while True:

            try:

                message_byte = await pubsub.get_message()

                params = await get_published_messages.get_redis_message(message_byte)

                data, message_channel = params["data"], params["channel"]

                if my_trade_receiving_channel in message_channel:

                    for trade in data:

                        instrument_name: str = trade["instrument_name"]
                        currency: str = trade["fee_currency"]
                        currency_lower: str = currency.lower()

                        archive_db_table = f"my_trades_all_{currency_lower}_json"

                        query_trades = (
                            f"SELECT * FROM  v_{currency_lower}_trading_active"
                        )

                        my_trades_currency_all_transactions: list = (
                            await db_mgt.executing_query_with_return(query_trades)
                        )

                        my_trades_currency_all: list = (
                            []
                            if my_trades_currency_all_transactions == 0
                            else [
                                o
                                for o in my_trades_currency_all_transactions
                                if o["instrument_name"]
                                in [
                                    o["instrument_name"]
                                    for o in instrument_attributes_futures_all
                                ]
                            ]
                        )

                        # handling transactions with no label
                        await labelling_blank_labels(
                            instrument_name,
                            my_trades_currency_all_transactions,
                            archive_db_table,
                        )

                        duplicated_trade_id_transactions = (
                            await db_mgt.querying_duplicated_transactions(
                                archive_db_table, "trade_id"
                            )
                        )

                        if duplicated_trade_id_transactions:

                            log.critical(
                                f"duplicated_trade_id_transactions {duplicated_trade_id_transactions}"
                            )

                            ids = [o["id"] for o in duplicated_trade_id_transactions]

                            for id in ids:
                                await db_mgt.deleting_row(
                                    archive_db_table,
                                    "databases/trading.sqlite3",
                                    "id",
                                    "=",
                                    id,
                                )

                if (
                    sqlite_updating_channel in message_channel
                    or portfolio_channel in message_channel
                ):

                    for currency in currencies:

                        currency_lower: str = currency.lower()

                        archive_db_table = f"my_trades_all_{currency_lower}_json"

                        query_trades = (
                            f"SELECT * FROM  v_{currency_lower}_trading_active"
                        )

                        my_trades_currency_all_transactions: list = (
                            await db_mgt.executing_query_with_return(query_trades)
                        )

                        my_trades_currency_all: list = (
                            []
                            if my_trades_currency_all_transactions == 0
                            else [
                                o
                                for o in my_trades_currency_all_transactions
                                if o["instrument_name"]
                                in [
                                    o["instrument_name"]
                                    for o in instrument_attributes_futures_all
                                ]
                            ]
                        )

                        # handling transactions with no label
                        await labelling_blank_labels(
                            currency,
                            my_trades_currency_all_transactions,
                            archive_db_table,
                        )

                        duplicated_trade_id_transactions = (
                            await db_mgt.querying_duplicated_transactions(
                                archive_db_table, "trade_id"
                            )
                        )

                        if duplicated_trade_id_transactions:

                            log.critical(
                                f"duplicated_trade_id_transactions {duplicated_trade_id_transactions}"
                            )

                            ids = [o["id"] for o in duplicated_trade_id_transactions]

                            for id in ids:
                                await db_mgt.deleting_row(
                                    archive_db_table,
                                    "databases/trading.sqlite3",
                                    "id",
                                    "=",
                                    id,
                                )

                            my_trades_currency_all_transactions: list = (
                                await db_mgt.executing_query_with_return(query_trades)
                            )

                            my_trades_currency: list = [
                                o
                                for o in my_trades_currency_all
                                if o["label"] is not None
                            ]

                            server_time = time_mod.get_now_unix_time()

                            #                            break

                            for strategy in active_strategies:

                                strategy_params = [
                                    o
                                    for o in strategy_attributes
                                    if o["strategy_label"] == strategy
                                ][0]

                                my_trades_currency_strategy = [
                                    o
                                    for o in my_trades_currency
                                    if strategy in (o["label"])
                                ]

                                if "futureSpread" in strategy:

                                    my_trades_currency_strategy_labels: list = [
                                        o["label"] for o in my_trades_currency_strategy
                                    ]

                                    # get labels from active trades
                                    labels = str_mod.remove_redundant_elements(
                                        my_trades_currency_strategy_labels
                                    )

                                    filter = "label"

                                    pairing_label = (
                                        await reassigning_labels.pairing_single_label(
                                            strategy_attributes,
                                            archive_db_table,
                                            my_trades_currency_strategy,
                                            server_time,
                                        )
                                    )

                                    if pairing_label:

                                        log.error(f"pairing_label {pairing_label}")

                                        cancellable_strategies = [
                                            o["strategy_label"]
                                            for o in strategy_attributes
                                            if o["cancellable"] == True
                                        ]

                                        await cancelling_active_orders.cancel_the_cancellables(
                                            client_id,
                                            client_secret,
                                            order_db_table,
                                            currency,
                                            cancellable_strategies,
                                        )

                                    #! closing active trades
                                    for label in labels:

                                        label_integer: int = str_mod.parsing_label(
                                            label
                                        )["int"]
                                        selected_transaction = [
                                            o
                                            for o in my_trades_currency_strategy
                                            if str(label_integer) in o["label"]
                                        ]

                                        selected_transaction_amount = [
                                            o["amount"] for o in selected_transaction
                                        ]

                                        sum_selected_transaction = sum(
                                            selected_transaction_amount
                                        )
                                        len_selected_transaction = len(
                                            selected_transaction_amount
                                        )

                                        #! closing combo auto trading
                                        if "Auto" in label:

                                            if sum_selected_transaction == 0:

                                                abnormal_transaction = [
                                                    o
                                                    for o in selected_transaction
                                                    if "closed" in o["label"]
                                                ]

                                            else:

                                                new_label = (
                                                    f"futureSpread-open-{label_integer}"
                                                )

                                                await db_mgt.update_status_data(
                                                    archive_db_table,
                                                    "label",
                                                    filter,
                                                    label,
                                                    new_label,
                                                    "=",
                                                )

                                                log.debug("renaming combo Auto done")

                                                await cancelling_active_orders.cancel_the_cancellables(
                                                    client_id,
                                                    client_secret,
                                                    order_db_table,
                                                    currency,
                                                    cancellable_strategies,
                                                )

                                                # break

                                        #! renaming combo auto trading
                                        else:

                                            if sum_selected_transaction == 0:

                                                if "open" in label:
                                                    new_label = f"futureSpreadAuto-open-{label_integer}"

                                                if "closed" in label:
                                                    new_label = f"futureSpreadAuto-closed-{label_integer}"

                                                await db_mgt.update_status_data(
                                                    archive_db_table,
                                                    "label",
                                                    filter,
                                                    label,
                                                    new_label,
                                                    "=",
                                                )

                                                # break

                                            #! closing unpaired transactions
                                            else:

                                                if len_selected_transaction != 1:

                                                    selected_transaction_trade_id = (
                                                        [
                                                            o["trade_id"]
                                                            for o in selected_transaction
                                                        ]
                                                    )[0]

                                                    filter = "trade_id"

                                                    if "open" in label:
                                                        new_label = f"futureSpread-open-{server_time}"

                                                    if "closed" in label:
                                                        new_label = f"futureSpread-closed-{server_time}"

                                                    await db_mgt.update_status_data(
                                                        archive_db_table,
                                                        "label",
                                                        filter,
                                                        selected_transaction_trade_id,
                                                        new_label,
                                                        "=",
                                                    )

                                                    # break

                                                else:

                                                    if "closed" not in label:
                                                        pass

                                if "hedgingSpot" in strategy:

                                    pass

            except Exception as error:

                await error_handling.parse_error_message_with_redis(
                    client_redis,
                    error,
                )

                continue

            finally:
                await asyncio.sleep(0.001)

    except Exception as error:

        await error_handling.parse_error_message_with_redis(
            client_redis,
            error,
        )


def get_settlement_period(strategy_attributes: list) -> list:

    return str_mod.remove_redundant_elements(
        str_mod.remove_double_brackets_in_list(
            [o["settlement_period"] for o in strategy_attributes]
        )
    )


def reading_from_pkl_data(end_point, currency, status: str = None) -> dict:
    """ """

    path: str = system_tools.provide_path_for_file(end_point, currency, status)
    return pickling.read_data(path)


async def update_db_pkl(path: str, data_orders: dict, currency: str) -> None:

    my_path_portfolio = system_tools.provide_path_for_file(path, currency)

    if currency_inline_with_database_address(currency, my_path_portfolio):

        pickling.replace_data(my_path_portfolio, data_orders)


def currency_inline_with_database_address(
    currency: str,
    database_address: str,
) -> bool:
    return currency.lower() in str(database_address)


async def labelling_blank_labels(
    instrument_name: str,
    my_trades_currency_active: list,
    archive_db_table: str,
) -> None:

    my_trades_currency_active_with_blanks = (
        []
        if not my_trades_currency_active
        else [o for o in my_trades_currency_active if o["label"] is None]
    )

    if my_trades_currency_active_with_blanks:
        column_trade: str = (
            "id",
            "instrument_name",
            "data",
            "label",
            "trade_id",
        )

        my_trades_currency_archive: list = (
            await db_mgt.executing_query_based_on_currency_or_instrument_and_strategy(
                archive_db_table, instrument_name, "all", "all", column_trade
            )
        )

        my_trades_currency_active_with_blanks = [
            o for o in my_trades_currency_archive if o["label"] is None
        ]

        my_trades_archive_instrument_id = [
            o["trade_id"] for o in my_trades_currency_active_with_blanks
        ]

        if my_trades_archive_instrument_id:
            for trade_id in my_trades_archive_instrument_id:

                transaction = str_mod.parsing_sqlite_json_output(
                    [
                        o["data"]
                        for o in my_trades_currency_active_with_blanks
                        if trade_id == o["trade_id"]
                    ]
                )[0]

                log.warning(f"transaction {transaction}")

                label_open: str = get_custom_label(transaction)

                where_filter = "trade_id"

                await db_mgt.update_status_data(
                    archive_db_table,
                    "label",
                    where_filter,
                    trade_id,
                    label_open,
                    "=",
                )


def get_custom_label(transaction: list) -> str:

    side = transaction["direction"]
    side_label = "Short" if side == "sell" else "Long"

    try:
        last_update = transaction["timestamp"]
    except:
        try:
            last_update = transaction["last_update_timestamp"]
        except:
            last_update = transaction["creation_timestamp"]

    return f"custom{side_label.title()}-open-{last_update}"



# -*- coding: utf-8 -*-

import asyncio
import orjson
import traceback

from loguru import logger as log

from streaming_helper.utilities import string_modification as str_mod, template


def parse_error_message(
    error: str,
    message: str = None,
) -> str:
    """

    Capture & emit error message

    Args:
        message (str): error message

    Returns:
        trace back error message

    """
    info = f"{traceback.format_exception(error)}"

    if message != None:
        info = f"{message} {traceback.format_exception(error)}"

    log.critical(f"{info}")

    return info


async def parse_error_message_with_redis(
    client_redis: object,
    error: str,
    message: str = None,
) -> None:
    """ """

    channel = "error"

    try:

        result: dict = template.redis_message_template()

        info = parse_error_message(
            error,
            message,
        )

        pub_message = dict(
            data=info,
        )

        result["params"].update({"channel": channel})
        result["params"].update(pub_message)

        # publishing message
        await client_redis.publish(
            channel,
            orjson.dumps(result),
        )

    except Exception as error:

        parse_error_message(
            error,
            message,
        )



# -*- coding: utf-8 -*-

# built ins
import asyncio

# user defined formula
from streaming_helper.restful_api.telegram import (
    end_point_params_template as end_point_telegram,
)
from streaming_helper.channel_management import get_published_messages
from streaming_helper.utilities import system_tools


async def telegram_messaging(
    client_redis: object,
    client_id: str,
    client_secret: str,
    redis_channels: list,
) -> None:
    """ """

    try:

        # connecting to redis pubsub
        pubsub: object = client_redis.pubsub()

        error_channel: str = "error"

        # get redis channels
        order_rest_channel: str = redis_channels["order_rest"]
        my_trade_receiving_channel: str = redis_channels["my_trade_receiving"]
        order_update_channel: str = redis_channels["order_cache_updating"]
        portfolio_channel: str = redis_channels["portfolio"]
        sqlite_updating_channel: str = redis_channels["sqlite_record_updating"]
        sub_account_cached_channel: str = redis_channels["sub_account_cache_updating"]
        # prepare channels placeholders
        channels = [error_channel, my_trade_receiving_channel]

        # subscribe to channels
        [await pubsub.subscribe(o) for o in channels]

        while True:

            try:

                message_byte = await pubsub.get_message()

                params = await get_published_messages.get_redis_message(message_byte)

                data = params["data"]

                message_channel = params["channel"]

                if error_channel in message_channel:

                    await end_point_telegram.send_message(
                        client_id,
                        client_secret,
                        data,
                    )

                if my_trade_receiving_channel in message_channel:

                    for trade in data:

                        if trade:
                            instrument_name = trade["instrument_name"]
                            amount = trade["amount"]
                            price = trade["price"]
                            direction = trade["direction"]

                            text = f"Trade: {direction} {amount} of {instrument_name}  @ {price} "

                            await end_point_telegram.send_message(
                                client_id,
                                client_secret,
                                text,
                            )

            except Exception as error:

                await system_tools.parse_error_message_with_redis(
                    client_redis,
                    error,
                )

                continue

            finally:
                await asyncio.sleep(0.001)

    except Exception as error:

        await system_tools.parse_error_message_with_redis(
            client_redis,
            error,
        )



# -*- coding: utf-8 -*-

from streaming_helper.restful_api import connector


def get_basic_https() -> str:
    return f"https://api.telegram.org/bot"


def message_end_point(
    bot_token: str,
    bot_chatID: str,
    bot_message: str,
) -> str:

    return (
        bot_token
        + ("/sendMessage?chat_id=")
        + bot_chatID
        + ("&parse_mode=HTML&text=")
        + str(bot_message)
    )


async def send_message(
    bot_token: str,
    bot_chatID: str,
    bot_message: str,
) -> None:
    return await connector.get_connected(
        get_basic_https(),
        message_end_point(bot_token, bot_chatID, bot_message),
    )



# -*- coding: utf-8 -*-


def remove_double_brackets_in_list(data: list) -> list:
    """_summary_

    Args:
        data (list): instance: [
                                ['BTC-30AUG24', 'BTC-6SEP24', 'BTC-27SEP24', 'BTC-27DEC24',
                                'BTC-28MAR25', 'BTC-27JUN25', 'BTC-PERPETUAL'
                                ],
                                ['ETH-30AUG24', 'ETH-6SEP24', 'ETH-27SEP24', 'ETH-27DEC24',
                                'ETH-28MAR25', 'ETH-27JUN25', 'ETH-PERPETUAL'
                                ]
                                ]

    Returns:
        list: _description_

    Reference:
        https://stackoverflow.com/questions/952914/how-do-i-make-a-flat-list-out-of-a-list-of-lists
    """
    return [o for os in data for o in os]


def get_duplicated_elements(data: list) -> list:
    """

    Args:
        data (list)

    Returns:
        list:

    Example:
        data_original = ['A', 'A', 'B', 'B', 'B', 'C']
        data_cleaned = ['A','B']

    Reference:
        https://www.geeksforgeeks.org/python-program-print-duplicates-list-integers/
        https://medium.com/@ryan_forrester_/finding-duplicates-in-python-lists-a-complete-guide-6092b34ac375

    """

    return list(set([x for x in data if data.count(x) > 1]))


def remove_redundant_elements(data: list) -> list:
    """
    Remove redundant items in a list

    Args:
        data (list)

    Returns:
        list:

    Example:
        data_original = ['A', 'A', 'B', 'B', 'B', 'C']
        data_cleaned = ['A','B','C']

    Reference:
        1. https://stackoverflow.com/questions/9427163/remove-duplicate-dict-in-list-in-python
        2. https://python.plainenglish.io/how-to-remove-duplicate-elements-from-lists-without-using-sets-in-python-5796e93e6d43
    """

    # Create an empty list
    result = []

    # Check if the data is a list and not empty
    if isinstance(data, list) and data != []:
        try:
            # Ref 1
            result = list({frozenset(item.items()): item for item in data}.values())

        except:
            # Ref 2
            result = list(dict.fromkeys(data))

    return result


def find_non_repeatable_elements(
    data1: list,
    data2: list,
) -> list:
    """

    Comparing two lists and picking non-repeatable items between them

    Args:
        data (list)  and its subset (list) for comparation

    Returns:
        list

    Example:
        data_original = [1, 2, 3, 4, 5] # all data
        data_redundant = [2, 4] # subset of all data
        data_cleaned =  [1, 3, 5]

    Reference:
        https://stackoverflow.com/questions/45098206/unique-values-between-2-lists

    """
    return [i for i in data1 if i not in data2]


def get_unique_elements(
    data1: list,
    data2: list,
) -> list:
    """

    Comparing two lists and picking only unique items between them

    Args:
        data (list)  and its subset (list) for comparation

    Returns:
        list

    Example:
        data1 = [4, 5, 6] # all data
        data_redundant = [1, 2, 3, 4, 5] # subset of all data
        data_cleaned =  [6]

    Reference:
        https://stackoverflow.com/questions/45098206/unique-values-between-2-lists

    """
    return list(set(data1).difference(data2))


def remove_dict_elements(
    original: dict,
    item_to_remove: str,
) -> str:
    """ """

    return {i: original[i] for i in original if i != item_to_remove}


def remove_list_elements(
    original: dict,
    item_to_remove: str,
) -> str:
    """
    https://stackoverflow.com/questions/13254241/removing-key-values-pairs-from-a-list-of-dictionaries
    """

    return [
        {key: value for key, value in dict.items() if key != item_to_remove}
        for dict in original
    ]


def extract_currency_from_text(words: str) -> str:
    """

    some variables:
    chart.trades.BTC-PERPETUAL.1
    incremental_ticker.BTC-4OCT24
    """

    if "." in words:
        filter1 = (words.partition(".")[2]).lower()

        if "." in filter1:
            filter1 = (filter1.partition(".")[2]).lower()

            if "chart.trades" in words:
                filter1 = (words.partition(".")[2]).lower()

            if "." in filter1:
                filter1 = (filter1.partition(".")[2]).lower()

                if "." in filter1:
                    filter1 = (filter1.partition(".")[0]).lower()

    else:
        filter1 = (words.partition(".")[0]).lower()

    return (filter1.partition("-")[0]).lower()


def remove_apostrophes_from_json(json_load: list) -> int:
    """ """
    import ast

    return [ast.literal_eval(str(i)) for i in json_load]


def parsing_sqlite_json_output(json_load: list) -> int:
    """
    parsing_sqlite_json_output

    References:
        https://stackoverflow.com/questions/46991650/remove-quotes-from-list-of-dictionaries-in-python
        https://stackoverflow.com/questions/14611352/malformed-string-valueerror-ast-literal-eval-with-string-representation-of-tup
    """

    try:

        result_json = [
            i.replace(":false", ":False")
            .replace(":true", ":True")
            .replace(":null", ":None")
            for i in json_load
        ]
        # print (f'result_json {[ast.literal_eval(str(i)) for i in result_json]}')
        return remove_apostrophes_from_json(result_json)

    except:
        return []


def parsing_redis_market_json_output(json_load: list) -> int:
    """
    parsing_sqlite_json_output

    References:
        https://stackoverflow.com/questions/46991650/remove-quotes-from-list-of-dictionaries-in-python
        https://stackoverflow.com/questions/14611352/malformed-string-valueerror-ast-literal-eval-with-string-representation-of-tup
    """

    try:

        result_json = [
            i.replace(":false", ":False").replace(":true", ":True") for i in json_load
        ]
        # print (f'result_json {[ast.literal_eval(str(i)) for i in result_json]}')
        return remove_apostrophes_from_json(result_json)

    except:
        return []


def get_strings_before_character(
    label: str,
    character: str = "-",
    character_place: int = [0, 2],
) -> str:
    """

    Get strings before a character

    Args:
        label (str)
        character (str)
        character_place (list (default)/int)

    Returns:
        str

    Example:
        data_original = 'hedgingSpot-open-1671189554374' become 'hedgingSpot'

    Reference:
        https://stackoverflow.com/questions/32682199/how-to-get-string-before-hyphen
    """

    if isinstance(character_place, list):
        splitted1 = label.split(character)[character_place[0]]
        splitted2 = label.split(character)[character_place[1]]
        splitted = f"{splitted1}-{splitted2}"
    else:
        splitted = label.split(character)[character_place]

    return splitted


def extract_integers_from_text(words: list) -> int:
    """
    Extracting integers from label text. More general than get integer in parsing label function
    """

    words_to_str = str(
        words
    )  # ensuring if integer used as argument, will be returned as itself

    return int("".join([o for o in words_to_str if o.isdigit()]))


def extract_integers_aggregation_from_text(
    identifier,
    aggregator,
    words: list,
) -> int:
    """
    identifier: id: trade/order/etc
    aggregator: min, max, len
    """

    return aggregator([extract_integers_from_text(o[f"{identifier}"]) for o in words])


def parsing_label(
    label: str,
    integer: int = None,
) -> dict:
    """

    Args:
        label (str)

    Returns:
        dict

    Example:
        'hedgingSpot-open-1671189554374'
        main: 'hedgingSpot'
        super_main: 'hedgingSpot'
        int = 1671189554374
        transaction_status:'hedgingSpot-open'
        transaction_net:'hedgingSpot-1671189554374'

        'every5mtestLong-open-1681617021717'
        main: 'every5mtestLong'
        super_main: 'every5mtest'
        int = 1681617021717
        transaction_status:'every5mtestLong-open''
        transaction_net:'every5mtestLong-1681617021717''

    """
    try:
        try:
            get_integer = get_strings_before_character(label, "-", 2)
        except:
            get_integer = get_strings_before_character(label, "-", 1)
    except:
        get_integer = None

    try:
        status = get_strings_before_character(label, "-", [0, 1])
    except:
        status = None

    try:
        net = get_strings_before_character(label)
    except:
        net = None

    try:
        main = get_strings_before_character(label, "-", 0)
    except:
        main = None

    try:
        side = ["Short", "Long"]
        super_main = [main.replace(o, "") for o in side if o in main]
    except:
        super_main = None

    try:
        closed_to_open = f"{main}-open-{get_integer}"

    except:
        closed_to_open = None

    try:
        if "Short" in main:
            flip = main.replace("Short", "Long")

        if "Long" in main:
            flip = main.replace("Long", "Short")
        flipping_closed = f"{flip}-open-{integer}"
    except:
        flipping_closed = None

    return {
        # "super_main":  bool([o not in main for o in side]),
        "super_main": (
            None
            if super_main == None
            else (main if all([o not in main for o in side]) else super_main[0])
        ),
        "main": main,
        "int": get_integer,
        "transaction_status": status,
        "transaction_net": net,
        "flipping_closed": flipping_closed,
        "closed_to_open": closed_to_open,
    }


def transform_nested_dict_to_list(list_example) -> dict:
    """ """
    len_tick = len(list_example["volume"])

    my_list = []

    for k in range(len_tick):

        dict_result = dict(
            volume=list_example["volume"][k],
            tick=list_example["ticks"][k],
            open=list_example["open"][k],
            low=list_example["low"][k],
            high=list_example["high"][k],
            cost=list_example["cost"][k],
            close=list_example["close"][k],
        )

        my_list.append(dict_result)

    return my_list


def transform_nested_dict_to_list_ohlc(list_example) -> dict:
    """ """
    len_tick = len(list_example["open"])

    my_list = []

    for k in range(len_tick):

        dict_result = dict(
            volume=list_example["volume"][k],
            tick=list_example["ticks"][k],
            open=list_example["open"][k],
            high=list_example["high"][k],
            low=list_example["low"][k],
            cost=list_example["cost"][k],
            close=list_example["close"][k],
        )

        my_list.append(dict_result)

    return my_list


def filtering_list_with_missing_key(
    list_examples: list,
    missing_key: str = "label",
) -> dict:
    """
    https://stackoverflow.com/questions/34710571/can-i-use-a-list-comprehension-on-a-list-of-dictionaries-if-a-key-is-missing

    """
    return [o for o in list_examples if missing_key not in o]


def sorting_list(
    listing: list,
    item_reference: str = "price",
    is_reversed: bool = True,
) -> list:
    """
    https://sparkbyexamples.com/python/sort-list-of-dictionaries-by-value-in-python/

    Args:
        listing (list): _description_
        item_reference (str, optional): _description_. Defaults to "price".
        is_reversed (bool, optional): _description_. Defaults to True.
                                    True = from_highest_to_lowest
                                    False = from_lowest_to_highest

    Returns:
        list: _description_
    """
    import operator

    return sorted(
        listing,
        key=operator.itemgetter(item_reference),
        reverse=is_reversed,
    )


def hashing(
    timestamp: int,
    client_id: str,
    apiSecret: str,
) -> str:

    import hashlib
    import hmac
    from urllib.parse import urlencode

    payload = {"apiKey": client_id, "timestamp": timestamp}

    return hmac.new(
        apiSecret.encode("utf-8"),
        urlencode(payload).encode("utf-8"),
        hashlib.sha256,
    ).hexdigest()


def convert_to_bytes(
    input_string: str,
    encoding: str = "utf-8",
) -> bytes:
    """
    https://www.w3resource.com/python-exercises/extended-data-types/python-extended-data-types-bytes-bytearrays-exercise-1.php

    encodings = [
        "utf-8",
        "utf-16",
        "ascii"
        ]

    """
    try:
        encoded_bytes = input_string.encode(encoding)
        return encoded_bytes

    except Exception as e:
        ("Error:", e)
        return None


def labelling(
    order: str,
    strategy: str,
    id_strategy: int = None,
) -> str:
    """
    labelling based on  strategy and unix time at order is made
    """

    id_unix_time = get_now_unix_time() if id_strategy is None else id_strategy

    return (
        (f"{strategy}-{order}-{id_unix_time}")
        if id_strategy is None
        else (f"{id_strategy}")
    )
